/*
 *
 * Airfinder - Helps locating a specific wireless MAC address
 * 			   (works better with a directionnal antenna)
 *
 *  Based on aircrack (made by Christophe Devine).
 *
 *  Copyright (C) 2005  Mister_X
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/if_ether.h>
#include <netpacket/packet.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <net/if.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#define ARPHRD_IEEE80211 801
#define ARPHRD_IEEE80211_PRISM 802

#define REFRESH_TIMEOUT 200000

#define S_SIZE 20

const char * usage =
"Airfinder v1.0beta\nHelps locating a specific wireless MAC address"
"\n(works better with a directionnal antenna)\n\nUsage: "
"airfinder interface channel MAC\n\n";

int main( int argc, char *argv[] )
{
	int n,i;
	int power;					 /* strength of signal if Prism2 header  */
	int fd_raw_in;				 /* file descriptor of the raw socket    */
	int caplen;					 /* length of captured packet            */
	int arptype;				 /* hardware linktype (packet header)    */
	char * s, *t;
	int atheros,lastpower, clientfound,channel;
	unsigned char mac[6];

	unsigned char *buffer;
	unsigned char *h80211;

	struct ifreq ifr;
	struct packet_mreq mr;
	struct sockaddr_ll sll;
	struct timeval tv;

	fd_set rfds;

	/* initialize a bunch of variables */
	clientfound =  0;
	channel     =  0;
	power       = -1;
	lastpower   = -1;
	fd_raw_in   = -1;
	arptype     =  0;

	//Faire un getenv(COMSPEC) pr voir si c win => quit

	/* check the arguments */

	if( argc != 4)
	{
		printf( usage);
		return( 1 );
	}

	atheros = ! memcmp(argv[1],"ath",3);

	channel  = atoi( argv[2] );
	if (channel<1)
	{
		printf(usage);
		exit(1);
	}

	if (!strcmp(argv[3],"FF:FF:FF:FF:FF:FF"))
	{
		/* invalid MAC - Broadcast address*/
		printf(usage);
		exit(1);
	}
	t = argv[3];
	i = 0;
	while( sscanf(t, "%x", &n ) == 1 )
	{
		if( n < 0 || n > 255 )
		{
			printf(usage);
			return( 1 );
		}
		mac[i] = n;
		if( ++i == 6 ) break;
		if( ! ( t = strchr( t, ':' ) ) )
			break;
		++t;
	}
	if ( i != 6 )
	{
		/* invalid MAC */
		printf("Invalid MAC; i=%d\n",i);
		printf(usage);
		exit(1);
	}
	/* create the raw socket and drop privileges */
	fd_raw_in = socket( PF_PACKET, SOCK_RAW, htons( ETH_P_ALL ) );
	setuid( getuid() );
	/* reserve the buffer space */
	if( ( buffer = (unsigned char *) calloc( 1,65536 ) ) == NULL )
	{
		perror( "malloc failed" );
		return( 1 );
	}

	/* try to open the first argument */

	if( fd_raw_in < 0 )
	{
		perror( "socket(PF_PACKET) failed" );
		fprintf( stderr, "This program requires root privileges.\n" );
		return( 1 );
	}

	/* find the interface index */

	memset( &ifr, 0, sizeof( ifr ) );
	strncpy( ifr.ifr_name, argv[1], sizeof( ifr.ifr_name ) - 1 );

	if( ioctl( fd_raw_in, SIOCGIFINDEX, &ifr ) < 0 )
	{
		fprintf( stderr, "\"%s\" is not a network interface, exiting.\n", argv[1] );
		return( 1 );
	}
	memset( &sll, 0, sizeof( sll ) );
	sll.sll_family   = AF_PACKET;
	sll.sll_ifindex  = ifr.ifr_ifindex;
	sll.sll_protocol = htons( ETH_P_ALL );
	if( memcmp( argv[1], "wlan", 4 ) == 0 )
	{
		if( ! fork() )			 /* wlan-ng brain damage */
		{
			close( 0 ); close( 1 ); close( 2 ); chdir( "/" );
			execlp( "wlanctl-ng", "wlanctl-ng", argv[1],
				"lnxreq_ifstate", "ifstate=enable", NULL );
			exit( 1 );
		}
		wait( NULL );

		if( ! fork() )			 /* hostap card reset */
		{
			close( 0 ); close( 1 ); close( 2 ); chdir( "/" );
			execlp( "iwpriv", "iwpriv", argv[1], "reset", "1", NULL );
			exit( 1 );
		}
		wait( NULL );
	}

	/* try to get in Monitor mode */

	if( memcmp( argv[1], "wlan", 4 ) == 0 )
	{
		/* possibly wlan-ng */
		snprintf( s,  S_SIZE - 1, "channel=%d", channel );
		if( ! fork() )
		{
			close( 0 ); close( 1 ); close( 2 ); chdir( "/" );
			execlp( "wlanctl-ng", "wlanctl-ng", argv[1],
				"lnxreq_wlansniff", "enable=true",
				"prismheader=true", "wlanheader=false",
				"stripfcs=true", "keepwepflags=true", s, NULL );
			exit( 1 );
		}

		wait( NULL );
	}

	memset( s, 0, S_SIZE );
	snprintf( s,  S_SIZE - 1, "%d", channel );

	#if 0
	/* orinoco monitor mode not enabled as ipw2200 *
	 * goes dead when calling iwpriv eth1 monitor  */

	if( memcmp( argv[1], "eth", 3 ) == 0 )
	{
		if( ! fork() )
		{
			close( 0 ); close( 1 ); close( 2 ); chdir( "/" );
			execlp( "iwpriv", "iwpriv", argv[1],
				"monitor", "1", s, 0 );
			exit( 1 );
		}

		wait( NULL );
	}
	#endif

	if( ! fork() )
	{
		/* try the generic way */

		close( 0 ); close( 1 ); close( 2 ); chdir( "/" );
		execlp( "iwconfig", "iwconfig", argv[1], "mode",
			"Monitor", "channel", s, NULL );
		exit( 1 );
	}

	wait( NULL );

	/* make sure the interface is up */

	ifr.ifr_flags = IFF_UP | IFF_BROADCAST | IFF_RUNNING;

	if( ioctl( fd_raw_in, SIOCSIFFLAGS, &ifr ) < 0 )
	{
		perror( "ioctl(SIOCSIFFLAGS) failed" );
		return( 1 );
	}

	/* bind the raw socket to the interface */

	if( bind( fd_raw_in, (struct sockaddr *) &sll,
		sizeof( sll ) ) < 0 )
	{
		perror( "bind(ETH_P_ALL) failed" );
		return( 1 );
	}

	/* couple of iwprivs to enable the prism header */

	if( ! fork() )				 /* hostap */
	{
		close( 0 ); close( 1 ); close( 2 ); chdir( "/" );
		execlp( "iwpriv", "iwpriv", argv[1], "monitor_type", "1", NULL );
		exit( 1 );
	}
	wait( NULL );

	if( ! fork() )				 /* r8180 */
	{
		close( 0 ); close( 1 ); close( 2 ); chdir( "/" );
		execlp( "iwpriv", "iwpriv", argv[1], "prismhdr", "1", NULL );
		exit( 1 );
	}
	wait( NULL );

	if( ! fork() )				 /* prism54 */
	{
		close( 0 ); close( 1 ); close( 2 ); chdir( "/" );
		execlp( "iwpriv", "iwpriv", argv[1], "set_prismhdr", "1", NULL );
		exit( 1 );
	}
	wait( NULL );

	/* make sure the WEP key is off */

	if( ! fork() )
	{
		close( 0 ); close( 1 ); close( 2 ); chdir( "/" );
		execlp( "iwconfig", "iwconfig", argv[1], "key", "off", NULL );
		exit( 1 );
	}
	wait( NULL );

	/* Set channel */

	s = (char *) calloc(1,S_SIZE);
	if( memcmp( argv[1], "wlan", 4 ) == 0 )
	{
		memset( s, 0, S_SIZE );
		snprintf( s,  S_SIZE - 1, "channel=%d", channel );

		if( ! fork() )
		{
			close( 0 ); close( 1 ); close( 2 ); chdir( "/" );
			execlp( "wlanctl-ng", "wlanctl-ng", argv[1],
				"lnxreq_wlansniff", "enable=true",
				"prismheader=true", "wlanheader=false",
				"stripfcs=true", "keepwepflags=true", s, NULL );
			exit( 1 );
		}

		wait( NULL );
	}
	memset( s, 0, S_SIZE );
	snprintf( s,  S_SIZE - 1, "%d", channel );

	if( ! fork() )
	{
		close( 0 ); close( 1 ); close( 2 ); chdir( "/" );
		execlp( "iwconfig", "iwconfig", argv[1], "channel", s, NULL );
		exit( 1 );
	}

	wait( NULL );
	free(s);

	/* lookup the hardware type */

	if( ioctl( fd_raw_in, SIOCGIFHWADDR, &ifr ) < 0 )
	{
		perror( "ioctl(SIOCGIFHWADDR) failed" );
		return( 1 );
	}

	arptype = ifr.ifr_hwaddr.sa_family;

	if( arptype != ARPHRD_IEEE80211 &&
		arptype != ARPHRD_IEEE80211_PRISM )
	{
		if( arptype == 1 )
			fprintf( stderr, "\nARP linktype is set to 1 (Ethernet) " );
		else
			fprintf( stderr, "\nUnsupported hardware link type %4d ",
				arptype );

		fprintf( stderr, "- expected ARPHRD_IEEE80211\nor ARPHRD_IEEE8021"
			"1_PRISM instead.  Make sure RFMON is enabled:\n"
			"run 'ifconfig %s up; iwconfig %s mode Monitor "
			"channel <#>'\n\n", argv[1], argv[1] );
		return( 1 );
	}

	/* enable promiscuous mode */

	memset( &mr, 0, sizeof( mr ) );
	mr.mr_ifindex = sll.sll_ifindex;
	mr.mr_type    = PACKET_MR_PROMISC;

	if( setsockopt( fd_raw_in, SOL_PACKET, PACKET_ADD_MEMBERSHIP,
		&mr, sizeof( mr ) ) < 0 )
	{
		perror( "setsockopt(PACKET_MR_PROMISC) failed" );
		return( 1 );
	}

	system("clear");
	printf("Searching for %s...\n",argv[3]);
	while( 1 )
	{

		/* capture one packet */

		FD_ZERO( &rfds );
		FD_SET( fd_raw_in, &rfds );

		tv.tv_sec  = 0;
		tv.tv_usec = REFRESH_TIMEOUT;

		if( select( fd_raw_in + 1, &rfds, NULL, NULL, &tv ) < 0 )
		{
			if( errno == EINTR ) continue;
			perror( "select failed" );
			return( 1 );
		}

		if( ! FD_ISSET( fd_raw_in, &rfds ) )
			continue;

		/* one packet available for reading */

		memset( buffer, 0, 4096 );

		if( ( caplen = read( fd_raw_in, buffer, 65535 ) ) < 0 )
		{
			perror( "read failed" );
			return( 1 );
		}

		/* if device is an atheros, remove the FCS */

		if (atheros) caplen -= 4;

		/* skip the prism header if present */
		h80211 = buffer;
		if( arptype == ARPHRD_IEEE80211_PRISM )
		{
			if( buffer[7] == 0x40 )
			{
				n = 0x40;
				power = buffer[0x33];
			}
			else
			{
				n = *(int *)( buffer + 4 );
				if( *(int *)( buffer + 0x44 ) != 0 )
					power = *(int *)( buffer + 0x44 );
				else
					power = *(int *)( buffer + 0x5C );
				if( n <= 0 || n > 4096 )
					continue;
			}
			h80211 += n;
			caplen -= n;
		}
		if (!memcmp(h80211+10,mac,6))
		{
			system("clear");
			printf("%s found on channel %d\nSignal power: %d\n",argv[3],channel,power);
			clientfound=1;
			if (lastpower!=-1)
			{
				if (power>lastpower) puts("You get closer");
				else puts("You get farer");
			}
			lastpower=power;
			//lasttime = time(NULL);
		}
		else
		{
			if (!clientfound)
				if (!memcmp(h80211+4,mac,6))
			{
				system("clear");
				printf("Found someone communicating with %s on channel %d\n"
					"Signal power: %d\nStill searching...",argv[3],channel,power);
			}
		}
		/*
		else
		{
			//Dure -> si pas de paquets depuis 3 secondes -> client perdu
			if (time(NULL)-lasttime>3000) lastpower = -1;
		}
		*/
	}
	return 0;
}
